@@ -1,7 +1,7 @@
 # HiQ version 1.1
 #
 # Copyright (c) 2022, Oracle and/or its affiliates.
-# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
+# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
 #
 
 __author__ = None
@@ -34,11 +34,74 @@ from hiq.hiq_utils import (
 from hiq.jack import Jack
 from hiq.monkeyking import LogMonkeyKing
 from hiq.tree import Tree
-from hiq.utils import _check_overhead, _get_args_spec, is_hiqed
-from hiq.memory import get_memory_mb
+from hiq.utils import _check_overhead, _get_full_argspecs, is_hiqed
+from hiq.memory import get_memory_mb, total_gpu_memory_mb
 
 here = os.path.dirname(os.path.realpath(__file__))
 
+# lib, class, func, tag
+import inspect
+import types
+
+
+def parse_lib(lib_name, res=[]):
+    if len(res) > 100:
+        return res
+    md = hiq.mod(lib_name)
+    if md is None:
+        return None
+    p = [lib_name, "", "", ""]
+    l = [i for i in dir(md) if not i.startswith("_") or i == "__call__"]
+    for i in l:
+        try:
+            st = f"__m=type(md.{i})"
+            itree.exe(st, locals())
+            if (
+                locals()["__m"] == types.FunctionType
+                or locals()["__m"] == types.MethodType
+            ):
+                itree.exe(f"__m_len =len(inspect.getsourcelines(md.{i})[0])", locals())
+                if locals()["__m_len"] < 10:
+                    continue
+                p[2], p[3] = i, i
+                res.append(list(p))
+                if len(res) > 100:
+                    return res
+            elif locals()["__m"] == type:
+                p[1] = i
+                itree.exe(f"__dir_list=dir(md.{i})", locals())
+                fs = [
+                    i
+                    for i in locals()["__dir_list"]
+                    if not i.startswith("_") or i == "__call__"
+                ]
+                for j in fs:
+                    st = f"__m2=type(md.{i}.{j})"
+                    itree.exe(st, locals())
+                    if locals()["__m2"] == types.FunctionType:
+                        itree.exe(
+                            f"__m_len =len(inspect.getsourcelines(md.{i}.{j})[0])",
+                            locals(),
+                        )
+                        if locals()["__m_len"] < 10:
+                            continue
+                        p[2], p[3] = j, j
+                        res.append(list(p))
+                p[1] = ""
+            elif locals()["__m"] == types.ModuleType or "<class 'abc.ABCMeta'>" == str(
+                locals()["__m"]
+            ):
+                r = parse_lib(f"{lib_name}.{i}", res)
+                if r is not None:
+                    res += r
+                    if len(res) > 100:
+                        return res[:100]
+                print(st, ", mtype:", locals()["__m"])
+
+        except AttributeError:
+            continue
+    return res
+
 
 class HiQException(Exception):
     pass
@@ -365,7 +428,7 @@ class HiQBase(itree.ForestStats, LogMonkeyKing):
             HiQBase: the current object itself
         """
         if s.enabled:
-            return
+            return s
         try:
             if reset_trace:
                 s.reset()
@@ -449,9 +512,16 @@ class HiQBase(itree.ForestStats, LogMonkeyKing):
                 itree.exe(d, locals())
             else:
                 itree.exe(f"s.o_{tag} = m_", locals())
-                spec_with_default, spec_wo_default = _get_args_spec(
-                    inspect.getfullargspec(inspect.unwrap(locals()["m_"]))
-                )
+                spec_with_default, spec_wo_default = _get_full_argspecs(locals()["m_"])
+
+                if inspect.ismethod(locals()["m_"]):
+                    spec_with_default = spec_with_default[
+                        spec_with_default.find(",") + 1 :
+                    ]
+                    spec_wo_default = spec_wo_default[spec_wo_default.find(",") + 1 :]
+                    # spec_with_default += ",*args,**kwargs"
+                    # spec_wo_default += ",*args,**kwargs"
+
                 assert (
                     re.match(r"^[\w\.]+$", tag)
                     and ";" not in spec_with_default
@@ -462,7 +532,6 @@ class HiQBase(itree.ForestStats, LogMonkeyKing):
                 signature = spec_with_default
                 if spec_with_default:
                     signature += ","
-
                 if signature.startswith("**"):
                     signature = "s=s," + signature
                 elif ",**" in signature:
@@ -777,8 +846,8 @@ if __name__ == "__main__":
         return 1
 
     h1 = HiQLatency(fast_fail=1, hiq_id_func=get_tau_id_)
-    h1.tau[0] = {"hello": Tree(extra={}, tid='simon', queue_lmk=Queue())}
-    h1.tau[1] = {"hello": Tree(extra={}, tid='simon', queue_lmk=None)}
+    h1.tau[0] = {"hello": Tree(extra={}, tid="simon", queue_lmk=Queue())}
+    h1.tau[1] = {"hello": Tree(extra={}, tid="simon", queue_lmk=None)}
     h1.tau[2] = {"hello": 123}
 
     d = pickle.dumps(h1)
@@ -791,8 +860,8 @@ if __name__ == "__main__":
     # print(h1.tau)
     print(type(h2))
     print(h2.tau)
-    print(type(h1.tau[0]['hello']))
-    print(type(h2.tau[0]['hello']))
-    print(h1.tau[0]['hello'].queue_lmk)
-    print(h2.tau[0]['hello'].queue_lmk)
+    print(type(h1.tau[0]["hello"]))
+    print(type(h2.tau[0]["hello"]))
+    print(h1.tau[0]["hello"].queue_lmk)
+    print(h2.tau[0]["hello"].queue_lmk)
     h2.show()
